"""
Sequence
=========

Generate a sequence of scintillations.

Note that this implementation takes into account varying correlation time by updating
the impulse response, and not resampling.

"""
import numpy as np
from scipy.signal import resample, fftconvolve

import collections
from scintillations.common import *
from scintillations.common import _fluctuations_with_variance

#from turbulence.vonkarman import covariance_wind as _covariance_vonkarman_wind

def variance_gaussian(distance, wavenumber, scale, mean_mu_squared, include_saturation=False):
    """Variance of Gaussian fluctuations.

    :param distance: Distance.
    :param wavenumber: Wavenumber.
    :param scale: Correlation length
    :param mean_mu_squared: Mean mu squared.

    :param include_saturation: Whether to include log-amplitude saturation. In this case the variance is multiplied with the saturation factor, :func:`saturation_factor`.
    :returns: Variance

    .. math:: \\langle \\chi^2 \\rangle = \\langle S^2 \\rangle = \\frac{\\sqrt{\\pi}}{2} \\langle \\mu^2 \\rangle k^2 r L

    """
    variance = np.sqrt(np.pi)/2.0 * mean_mu_squared * wavenumber*wavenumber * distance * scale
    if include_saturation:
        variance *= saturation_factor(distance, wavenumber, scale, mean_mu_squared)
    return variance


def fluctuations_with_variance(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation=False):
    """Generate fluctuations with correct variance.
    """
    return _fluctuations_with_variance(variance_gaussian, fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation)


def covariance_gaussian(spatial_separation, distance, wavenumber, scale, mean_mu_squared):
    """Calculate the covariance of a Gaussian turbulence spectrum and spherical waves.

    See Daigle, 1987: equation 2 and 3.

    :param spatial_separation: Spatial separation.
    :param distance: Distance.
    :param wavenumber: Wavenumber.
    :param mean_mu_squared: Mean mu squared.
    :param scale: Outer length scale.
    :returns: Covariance

    .. math:: B{\\chi} (\\rho) = B{S}(\\rho) = \\frac{\\sqrt{\\pi}}{2} \\langle \\mu^2 \\rangle k^2 r L \\frac{\\Phi(\\rho/L) }{\\rho / L}

    """
    #covariance = 0.0
    #covariance += (spatial_separation!=0.0) * \
                  #np.nan_to_num( ( np.pi/4.0 * mean_mu_squared * (wavenumber*wavenumber) * \
                  #distance * scale * (erf(spatial_separation/scale) / \
                  #(spatial_separation/scale) ) ) )

    #covariance += (spatial_separation==0.0) * np.sqrt(np.pi)/2.0 * \
                  #mean_mu_squared * (wavenumber*wavenumber) * distance * scale
    cor = correlation_spherical_wave(spatial_separation, scale)
    var = variance_gaussian(distance, wavenumber, scale, mean_mu_squared)
    covariance = cor * var
    return covariance




def logamp_structure(logamp_a, logamp_b, axis=-1):
    """Structure function for log-amplitude fluctuations.

    See Daigle, equation 17.
    """
    return ((logamp_a-logamp_b)**2.0).mean(axis=axis)


def phase_structure(phase_a, phase_b, axis=-1):
    """Structure function for phase fluctuations.

    See Daigle, equation 18.

    The last term accounts for the fact that the mean phase difference of the signals may be nonzero.
    """
    return ((phase_a-phase_b)**2.0).mean(axis=axis) - ((phase_a-phase_b).mean(axis=axis))**2.0


def logamp_variance(logamp, axis=-1):
    """Logamp variance.

    See Daigle,  equation 19.
    """
    amp = np.exp(logamp)
    logamp_normalized = np.log(amp/amp.mean(axis=axis))
    return logamp_normalized.var(axis=axis)


def phase_variance(phase, axis=-1):
    """Phase variance.

    See Daigle, equation 20.

    The last term accounts for the fact that the mean phase difference of the signals may be nonzero.
    """
    return phase.var(axis=axis) - (phase-phase.mean(axis=axis)).mean(axis=axis)**2.0


def generate_gaussian_fluctuations_standard(nsamples, ntaps, fs, correlation_time, state=None, window=None):
    """Generate Gaussian fluctuations with variance 1.

    :param nsamples: Length of the sequence in samples.
    :param ntaps: Length of the filter used to shape the PSD of the sequence.
    :param fs: Sample frequency.
    :param correlation_time: Correlation time. Single value or array with a value for each time instance.
    :param state: State of random number generator.
    :param window: Window used in filter design.
    :returns: Fluctuations.

    This function generates fluctuations that are Gaussian distributed and have a variance of 1.

    """
    # Gaussian white noise
    state = state if state else np.random.RandomState()
    nsamples_total = nsamples + ntaps - 1
    noise = state.randn(nsamples_total)

    # Time shifts for sampling temporal correlation function.
    times = tau(ntaps, fs)

    # Whether our correlation time is time-variant or not.
    # This is an optimization step.
    if isinstance(correlation_time, collections.Iterable):
        if len(set(correlation_time)) > 1:  # Unique values
            time_variant = True
        else:   # Sequence with same value
            correlation_time = correlation_time[0]
            time_variant = False
    else:   # Single value
        time_variant = False;

    if time_variant:
        # We need to convolve each sample with a unique impulse response.
        fluctuations = np.empty(nsamples)
        for i in range(nsamples):
            noise_block = noise[i:i+ntaps]
            correlation = correlation_spherical_wave(times, correlation_time[i])
            ir = impulse_response_fluctuations(correlation, ntaps, window=window)
            fluctuations[i] = fftconvolve(noise_block, ir, mode='valid')

    else:
        # If time-invariant, we can perform a single convolution.
        correlation = correlation_spherical_wave(times, correlation_time)
        ir = impulse_response_fluctuations(correlation, ntaps, window=window)
        fluctuations = fftconvolve(noise, ir, mode='valid')

    # The filtering process adjusts the spectrum of the sequence and thereby the variance.
    # One might want to normalize with the standard deviation to take this into account.
    # However, the expected value of the sequence is still one, and therefore
    # normalization should not be done.
    #fluctuations /= fluctuations.std()

    return fluctuations




def generate_fluctuations_logamp_and_phase(nsamples, fs, ntaps, correlation_length, speed, frequency, soundspeed,
                                           distance, mean_mu_squared, include_saturation=True, state=None, window=None):
    """Generate logamp and phase fluctuations.

    :returns: Logamp :math:`\chi(t)` and phase :math:`S(t)` fluctuations.

    """
    correlation_length = np.atleast_1d(correlation_length)
    speed = np.atleast_1d(speed)
    correlation_time = correlation_length / speed
    mean_mu_squared = np.atleast_1d(mean_mu_squared)
    distance = np.atleast_1d(distance)

    wavenumber = 2.*np.pi*frequency / soundspeed
    # Fluctuations with variance 1.
    fluctuations = generate_gaussian_fluctuations_standard(nsamples, ntaps, fs, correlation_time, state=state, window=window)

    # Log-amplitude and phase fluctuations with correct variance.
    logamp = fluctuations_with_variance(fluctuations[:,None], wavenumber, distance[:,None], correlation_length[:,None], mean_mu_squared[:,None], include_saturation)
    phase =  fluctuations_with_variance(fluctuations[:,None], wavenumber, distance[:,None], correlation_length[:,None], mean_mu_squared[:,None], False)

    return np.squeeze(logamp), np.squeeze(phase)


def generate_fluctuations_spectra_and_delay(nsamples, fs, ntaps, correlation_length, speed, frequency, soundspeed,
                                            distance, mean_mu_squared, include_saturation=True, state=None, window=None):
    """Generate fluctuations represented by a time-variant magnitude-only spectrum and a variable delay.

    :returns: Magnitude-only spectra as function of time, and delay as function of time
    .. seealso:: :func:`generate_complex_fluctuations`
    """
    correlation_length = np.atleast_1d(correlation_length)
    speed = np.atleast_1d(speed)
    correlation_time = correlation_length / speed
    mean_mu_squared = np.atleast_1d(mean_mu_squared)
    distance = np.atleast_1d(distance)

    wavenumber = 2.*np.pi*frequency / soundspeed
    # Fluctuations with variance 1.
    fluctuations = generate_gaussian_fluctuations_standard(nsamples, ntaps, fs, correlation_time, state=state, window=window)

    # Log-amplitude and phase fluctuations with correct variance.
    logamp = fluctuations_with_variance(fluctuations[:,None], wavenumber, distance[:,None], correlation_length[:,None], mean_mu_squared[:,None], include_saturation=include_saturation)
    # Phase fluctuations with correct variance for frequency=1
    phase =  fluctuations_with_variance(fluctuations[:,None], 2.*np.pi*1.0/soundspeed, distance[:,None], correlation_length[:,None], mean_mu_squared[:,None], False)

    # Magnitude spectra
    spectra = np.squeeze(amplitude_fluctuations(logamp))

    # Variable delay
    delay = np.squeeze(delay_fluctuations(phase, fs, 1.0))

    return spectra, delay


def apply_log_amplitude(signal, log_amplitude):
    """Apply log-amplitude fluctuations.

    :param signal: Pressure signal.
    :param log_amplitude: Log-amplitude fluctuations.

    .. math:: p_m = p \\exp{\\chi}

    """
    return signal * amplitude_fluctuations(log_amplitude)


def apply_phase(signal, phase, frequency, fs):
    """Apply phase fluctuations.

    :param signal: Pressure signal.
    :param phase: Phase fluctuations.
    :param frequency: Frequency of tone.
    :param fs: Sample frequency.

    Phase fluctuations are applied through a resampling.

    """
    delay = delay_fluctuations(phase, frequency)
    signal = apply_delay(signal, delay, fs)
    return signal


def apply_fluctuations(signal, fs, frequency=None, log_amplitude=None, phase=None):
    """Apply log-amplitude and/or phase fluctuations.
    """
    if log_amplitude is not None:
        signal = apply_log_amplitude(signal, log_amplitude)
    if phase is not None:
        signal = apply_phase(signal, phase, frequency, fs)
    return signal


def apply_delay(signal, delay, fs):
    """Apply propagation delay fluctuations.

    :param signal: Signal
    :param delay: Propagation delay fluctuations.
    :param fs: Sample frequency
    :returns: Frequency-modulated signal.
    """

    k_r = np.arange(0, len(signal), 1)          # Create vector of indices
    k = k_r - delay * fs                      # Create vector of warped indices

    kf = np.floor(k).astype(int)       # Floor the warped indices. Convert to integers so we can use them as indices.
    dk = kf - k
    ko = np.copy(kf)
    kf[ko<0] = 0
    kf[ko+1>=len(ko)] = 0
    R = ( (1.0 + dk) * signal[kf] + (-dk) * signal[kf+1] ) * (ko >= 0) * (ko+1 < len(k)) #+ 0.0 * (kf<0)
    return R


def modulate_tone(signal, fs, frequency, fs_low, correlation_length, speed, distance, soundspeed, mean_mu_squared,
                  ntaps_corr=8192, window=None, include_saturation=False, state=None, include_amplitude=True,
                  include_phase=True):
    """Modulate tone.

    :param signal: Signal to modulate.
    :param fs: Sample frequency of signal.
    :param fs_low: Sample frequency at which to generate the scintillations.

    .. note:: Time-dependent values like speed and correlation length are given per block.
    """

    nsamples = len(signal)
    nsamples_low = int(np.ceil(nsamples * fs_low / fs))
    times = np.arange(nsamples)/fs
    times_low = np.arange(nsamples_low)/fs_low

    logamp, phase = generate_fluctuations_logamp_and_phase(nsamples_low, fs_low, ntaps_corr, correlation_length, speed, frequency,
                                                           soundspeed, distance, mean_mu_squared, include_saturation=include_saturation,
                                                           state=state, window=window)

    if include_amplitude:
        logamp = np.interp(times, times_low, logamp)
        signal = apply_log_amplitude(signal, logamp)

    if include_phase:
        delay = delay_fluctuations(phase, fs_low, frequency)
        delay = np.interp(times, times_low, delay)
        #delay = delay * fs_low #/ frequency # FIXME: Apparently we need this factor. WHY???!
        #phase = np.interp(times, times_low, phase)
        #delay = delay_fluctuations(phase, fs_low)

        signal = apply_delay(signal, delay, fs)

    return signal


def modulate_broadband(signal, fs, nblock, correlation_length, speed, distance, soundspeed, mean_mu_squared, ntaps_corr=8192,
             ntaps_spectra=128, window=None, include_saturation=False, state=None,
             include_amplitude=True, include_phase=True):
    """Modulate tone.
    """
    return NotImplemented
    fb = fs / nblock
    frequency = np.fft.rfftfreq(ntaps_spectra, 1./ fs)
    spectrum, delay = generate_fluctuations_spectra_and_delay(nsamples, fb, ntaps_corr, correlation_length, speed,
                                                                 frequency, soundspeed, distance, mean_mu_squared,
                                                                 include_saturation, state, window)
    if include_amplitude:
        return NotImplemented
        logamp = np.interp(times, times_low, logamp)
        signal = apply_log_amplitude(signal, logamp)

    if include_phase:
        delay = np.interp(times, times_low, delay)
        signal = apply_delay(signal, delay, fs)

    return signal

#def generate_gaussian_fluctuations(nsamples, ntaps, fs, correlation_length, speed, distance,
                                   #frequency, soundspeed, mean_mu_squared,
                                   #window=None, include_saturation=False,
                                   #state=None, factor=5.0):
    #"""Generate Gaussian fluctuations.

    #:param factor: To resolve spatial field you need a sufficient resolution.

    #This function generated log-amplitude and phase fluctuations with the correct variance for the specified frequency.

    #.. warning:: You likely do not want to use a window as it will dramatically alter the frequency response of the fluctuations.
    #.. seealso:: :func:`generate_gaussian_fluctuations_standard`
    #"""
    #correlation_time = correlation_length / speed
    ## Low resolution parameters
    #fs_low = factor / correlation_time
    #times = np.arange(nsamples)/fs
    ###correlation = correlation_spherical_wave(tau(ntaps, fs_low), correlation_time)
    #upsample_factor = fs / fs_low
    #nsamples_low = np.ceil(nsamples / upsample_factor)
    #times_low = np.arange(nsamples_low) / fs_low

    ## Modulation signal with variance 1
    #fluctuations = generate_gaussian_fluctuations_standard(nsamples_low, ntaps, fs_low, correlation_time, state, window)

    #wavenumber = 2.*np.pi*frequency / soundspeed

    ## Log-amplitude fluctuations with correct variance.
    #logamp = logamp_fluctuations(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared,
                                 #include_saturation=include_saturation)

    ## Phase fluctuations with correct variance.
    #phase = phase_fluctuations(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared)

    ## Upsampled modulation signals
    #logamp = np.interp(times, times_low, logamp)
    #phase = np.interp(times, times_low, phase)

    #return logamp, phase

#from scipy.signal import resample

#def upsample(signal, factor, method='fourier'):
    #"""Upsample signal with `factor`.

    #:param signal: Signal to be upsampled.
    #:param factor: Upsample factor.
    #:type factor: float
    #:returns: Upsampled signal.

    #.. seealso:: :func:`scipy.signal.resample`

    #"""
    #nsamples_low = len(signal)
    #nsamples_high = nsamples_low * np.ceil(factor)

    #if method=='fourier':
        #upsampled = resample(signal, nsamples_high)
    #elif method=='linear':
        #times_low = np.linspace(0., 1., nsamples_low, endpoint=False)
        #times_high = np.linspace(0., 1., nsamples_high, endpoint=False)
        #upsampled = np.interp(times_high, times_low, signal)
    #else:
        #raise ValueError("Invalid method: {}".format(method))

    #return upsampled


    #if nsamples_upsampled:
        #times = np.arange(nsamples)
        #fs_upsampled = nsamples_upsampled / nsamples * fs
        #times_upsampled = np.arange(nsamples_upsampled) / fs_upsampled
        #logamp = np.interp(times_upsampled, times, logamp)
        #phase = np.interp(times_upsampled, times, phase)



#def modulate(signal, fs, correlation_length, speed, distance, soundspeed, mean_mu_squared, ntaps=8192,
             #nfreqs=100, window=None, include_saturation=False, state=None, factor=5.0,
             #include_amplitude=True, include_phase=True, fs_low=None):
    #"""Apply modulations to `signal`.

    #:param signal: Signal to apply modulations to.
    #:param fs: Sample frequency of signal.
    #:param correlation_length: Correlation length.
    #:param speed: Speed.
    #:param distance: Distance.
    #:param soundspeed: Speed of sound.
    #:param mean_mu_squared: Variance of refractive-index.
    #:param ntaps: Amount of taps to use to sample the correlation function.
    #:param nfreqs: Amount of frequencies to calculate the logamp variances for.
    #:param window: Window to apply when designing filter.
    #:param include_saturation: Include logamp saturation.
    #:param state: State of the PRNG.
    #:param factor: Oversampling factor.
    #:param include_amplitude: Whether to include logamp modulations.
    #:param include_phase: Whether to include phase modulations.

    #"""
    #correlation_time = correlation_length / speed
    #if fs_low is None:
        #fs_low = np.max(5.0 / correlation_time)

    #times = np.arange(nsamples)/fs

    #upsample_factor = fs / fs_low
    #nsamples_low = np.ceil(nsamples / upsample_factor)
    #times_low = np.arange(nsamples_low) / fs_low

    #fluctuations = generate_complex_fluctuations




def covariance(covariance_func, **kwargs):
    if covariance_func == 'gaussian':
        return covariance_gaussian(kwargs['spatial_separation'], kwargs['distance'],
                                   kwargs['wavenumber'], kwargs['scale'], kwargs['mean_mu_squared'])
    elif covariance_func == 'vonkarman_wind':
        return _covariance_vonkarman_wind(kwargs['spatial_separation'], kwargs['distance'],
                                          kwargs['wavenumber'], kwargs['scale'], kwargs['soundspeed'],
                                          kwargs['wind_speed_variance'], kwargs['steps'], kwargs['initial'])
    else:
        raise ValueError("Unknown covariance function {}".format(covariance_func))


def get_covariance(covariance):

    if covariance == 'gaussian':
        def wrapped(**kwargs):
            return covariance_gaussian(kwargs['spatial_separation'], kwargs['distance'],
                                       kwargs['wavenumber'], kwargs['scale'], kwargs['mean_mu_squared'])
    elif covariance == 'vonkarman_wind':
        def wrapped(**kwargs):
            return _covariance_vonkarman_wind(kwargs['spatial_separation'], kwargs['distance'],
                                              kwargs['wavenumber'], kwargs['scale'], kwargs['soundspeed'],
                                              kwargs['wind_speed_variance'], kwargs['steps'], kwargs['initial'])
    else:
        raise ValueError("Covariance unavailable.")

    return wrapped

#def covariance_gaussian(**kwargs):
    #return covariance(spatial_separation, distance, wavenumber, scale, mean_mu_squared)

#def covariance_vonkarman_wind(**kwargs):
    #return _covariance_vonkarman_wind(spatial_separation, distance, wavenumber, scale, soundspeed, wind_speed_variance, steps, initial)

#COVARIANCES = {
        #'gaussian' : covariance,
        #'vonkarman_wind' : covariance_vonkarman_wind,
    #}


###def generate_fluctuations(nsamples, ntaps, fs, speed, distance,
                          ###frequency, soundspeed, scale, state=None,
                          ###window=None, model='gaussian', **kwargs):

    ###logging.debug("generate_fluctuations: covariance model {}".format(model))

    ###try:
        ###include_saturation = kwargs.pop('include_saturation')
    ###except KeyError:
        ###include_saturation = False

    #### Determine the covariance
    ###spatial_separation = tau(ntaps, fs) * speed
    ###wavenumber = 2.*np.pi*frequency / soundspeed

    ###cov = covariance(model, spatial_separation=spatial_separation,
                                 ###distance=distance,
                                 ###wavenumber=wavenumber,
                                 ###scale=scale, **kwargs)

    ####cov0 = covariance_func(spatial_separation=0.0,
                                 ####distance=distance,
                                 ####wavenumber=wavenumber,
                                 ####scale=scale, **kwargs)
    #### Create an impulse response using this covariance
    ###ir = impulse_response_fluctuations(cov, window=window)

    #### We need random numbers.
    ###state = state if state else np.random.RandomState()

    #### Calculate log-amplitude fluctuations
    ####noise = state.randn(samples*2-1)
    ####log_amplitude = fftconvolve(noise, ir, mode='valid')
    ###noise = state.randn(nsamples)
    ###log_amplitude = fftconvolve(noise, ir, mode='same')
    ####log_amplitude -= cov[0]


    ####log_amplitude -= (log_amplitude.mean() - logamp_variance(np.exp(log_amplitude)))

    #### Include log-amplitude saturation
    ###if include_saturation:
        ###if model == 'gaussian':
            ###mean_mu_squared = kwargs['mean_mu_squared']
            ###sat_distance = saturation_distance(mean_mu_squared, wavenumber, scale)
            ###log_amplitude *=  (np.sqrt( 1.0 / (1.0 + distance/sat_distance) ) )
        ###else:
            ###raise ValueError("Cannot include saturation for given covariance function.")

    #### Calculate phase fluctuations
    ####noise = state.randn(samples*2-1)
    ####phase = fftconvolve(noise, ir, mode='valid')
    ###noise = state.randn(nsamples)
    ###phase = fftconvolve(noise, ir, mode='same')

    ###return log_amplitude, phase

#def fluctuations_logamp(ir, noise):
    #log_amplitude = fftconvolve(noise, ir, mode='same')
    #return log_amplitude

#def fluctuations_phase(ir, noise):
    #phase = fftconvolve(noise, ir, mode='same')
    #return phase





#def logamp_variance(logamp, axis=-1):
    #amp = np.exp(logamp)
    #logamp_normalized = np.log(amp/amp.mean(axis=axis))
    #return logamp_normalized.var(axis=axis)

#def logamp_variance(amp):
    #"""Variance of log-amplitude fluctuations.

    #:param amp: Time-series of amplitude fluctuations, NOT log-amplitude.

    #See Daigle, 1983: equation 15, 16 and 19.
    #"""
    #return (np.log(amp/(amp.mean(axis=-1)[...,None]))**2.0).mean(axis=-1)
    ##return (( np.log(amp) - np.log(amp.mean(axis=-1) )[...,None])**2.0).mean(axis=-1)


#def generate_many_gaussian_fluctuations(samples, spatial_separation, distance, wavenumber,
                                        #mean_mu_squared, scale, window=np.hamming,
                                        #include_saturation=False, seed=None):
    #"""Generate time series of log-amplitude and phase fluctuations.

    #:param samples: Length of series of fluctuations.
    #:param spatial_separation: Spatial separation.
    #:param distance: Distance.
    #:param wavenumber: Wavenumber
    #:param mean_mu_squared: Mean mu squared.
    #:param scale: Outer length scale.
    #:param window: Window function.
    #:param include_saturation: Include saturation of log-amplitude.
    #:param seed: Seed.
    #:returns: Log-amplitude array and phase array.

    #This function performs better when many series need to be generated.

    #"""

    ## Calculate correlation
    ##B = (spatial_separation!=0.0) * np.nan_to_num( ( np.pi/4.0 * mean_mu_squared * (k*k)[:,None] * r[None,:] * L * (erf(spatial_separation/L) / (spatial_separation/L))[None,:] ) )
    ##B = (spatial_separation==0.0)[None,:] * (np.sqrt(np.pi)/2.0 * mean_mu_squared * (k*k)* r * L )[:,None]

    #spatial_separation = np.atleast_1d(spatial_separation)
    #distance = np.atleast_1d(distance)
    #wavenumber = np.atleast_1d(wavenumber)
    #mean_mu_squared = np.atleast_1d(mean_mu_squared)
    #scale = np.atleast_1d(scale)
    #covariance = covariance_gaussian(spatial_separation[None,:], distance[:,None],
                                          #wavenumber[:,None], mean_mu_squared[:,None], scale[:,None])

    #if covariance.ndim==2:
        #N = covariance.shape[-2]
    #elif covariance.ndim==1:
        #N = 1
    #else:
        #raise ValueError("Unsupported amount of dimensions.")

    ## Seed random numbers generator.
    #np.random.seed(seed)
    #n = samples * 2 - 1

    #ir = impulse_response_fluctuations(covariance, window=window)

    #noise = np.random.randn(N,n)
    #log_amplitude = fftconvolve1D(noise, ir, mode='valid') # Log-amplitude fluctuations
    #del noise

    #if include_saturation:
        #sat_distance = saturation_distance(mean_mu_squared, wavenumber, scale)
        #log_amplitude *=  (np.sqrt( 1.0 / (1.0 + distance/sat_distance) ) )[...,None]
        #del sat_distance

    #noise = np.random.randn(N,n)
    #phase = fftconvolve1D(noise, ir, mode='valid')           # Phase fluctuations

    #return log_amplitude, phase




#def gaussian_fluctuations_variances(samples, f0, fs, mean_mu_squared,
                                    #distance, scale,
                                    #spatial_separation, soundspeed,
                                    #include_saturation=True, state=None):
    #"""Calculate the variances of fluctuations in the time series of amplitude and phase fluctuations.

    #:param samples: Amount of samples to take.
    #:param f0: Frequency for which the fluctuations should be calculated.
    #:param fs: Sample frequency.
    #:param mean_mu_squared: Mean of refractive-index squared.
    #:param r: Distance.
    #:param L: Outer length scale.
    #:param rho: Spatial separation.
    #:param soundspeed: Speed of sound.
    #:param include_phase: Include phase fluctuations.
    #:param state: State of numpy random number generator.

    #"""
    #spatial_separation *= np.ones(samples)
    #wavenumber = 2.0 * np.pi * f0 / soundspeed
    #a, p = generate_gaussian_fluctuations(samples, spatial_separation, distance,
                                               #wavenumber, mean_mu_squared, scale,
                                               #include_saturation=include_saturation,
                                               #state=state)

    #return logamp_variance(np.exp(a)), phase_variance(p)


#def fluctuations_variance(signals, fs, N=None):
    #"""
    #Determine the variance in log-amplitude and phase by ensemble averaging.

    #:param signals: List of signals or samples.
    #:param fs: Sample frequency

    #The single-sided spectrum is calculated for each signal/sample.

    #The log-amplitude of the :math:`n`th sample is given by

    #.. math:: \\chi^2 = \\ln{\\frac{A_{n}}{A_{0}}}

    #where :math:`A_{n}` is the amplitude of sample :math:`n` and :math:`A_{0}` is the ensemble average

    #.. math:: A_{0} = \\frac{1}{N} \\sum_{n=1}^{N} \\chi_{n}^2

    #"""

    #s = np.array(signals) # Array of signals

    ##print s

    #f, fr = ir2fr(s, fs, N) # Single sided spectrum
    #amp = np.abs(fr)
    #phase = np.angle(fr)
    #logamp_squared_variance = (np.log(amp/amp.mean(axis=0))**2.0).mean(axis=0)
    #phase_squared_variance = ((phase - phase.mean(axis=0))**2.0).mean(axis=0)

    #return f, logamp_squared_variance, phase_squared_variance


#def plot_variance(frequency, logamp, phase):
    #"""
    #Plot variance.
    #"""
    #fig = plt.figure()
    #ax = fig.add_subplot(111)
    #ax.scatter(frequency, logamp, label=r"$\langle \chi^2 \rangle$", color='b')
    #ax.scatter(frequency, phase, label=r"$\langle S^2 \rangle$", color='r')
    #ax.set_xlim(100.0, frequency.max())
    #ax.set_ylim(0.001, 10.0)
    #ax.set_xscale('log')
    #ax.set_yscale('log')
    #ax.legend()
    #ax.grid()
    #ax.set_xlabel(r"$f$ in Hz")
    #ax.set_ylabel(r"$\langle X \rangle$")

    #return fig


#def _spatial_separation(A, B, C):
    #"""Spatial separation.

    #:param A: Source position as function of time.
    #:param B: Reference position, e.g. source position at t-1
    #:param C: Receiver position.

    #Each row is a sample and each column a spatial dimension.

    #"""
    #a = np.linalg.norm(B-C, axis=1)
    #b = np.linalg.norm(A-C, axis=1)
    #c = np.linalg.norm(A-B, axis=1)

    #gamma = np.arccos((a**2.0+b**2.0-c**2.0) / (2.0*a*b))
    #spatial_separation = 2.0 * b * np.sin(gamma/2.0)
    #L = b * np.cos(gamma/2.0)

    #return spatial_separation, L


#def transverse_speed(A, B, C, fs):
    #"""Transverse speed computed from three positions.

    #:param A: Source position as function of time.
    #:param B: Reference position, e.g. source position at t=0
    #:param C: Receiver position.
    #:param fs: Sample frequency

    #"""
    #rho, _ = _spatial_separation(A, B, C)

    #step_distance = np.linalg.norm(A-B, axis=-1)
    #speed = step_distance * fs
    #transverse = rho / step_distance * speed

    #return transverse, rho, step_distance






"""
Common
======

Expressions common to both :mod:`stream` and :mod:`sequence`.
"""

import numpy as np
from scipy.special import erf
from scipy.fftpack import dct


def saturation_distance(mean_mu_squared, wavenumber, scale):
    """Saturation distance according to Wenzel.

    :param mean_mu_squared: Mean mu squared.
    :param wavenumber: Wavenumber.
    :param scale: Outer length scale.

    See Daigle, 1987: equation 5

    .. math:: r_s = \\frac{1}{2 \\langle \mu^2 \\rangle k^2 L}

    """
    return 1.0 / (2.0 * mean_mu_squared * wavenumber*wavenumber * scale)


def saturation_factor(distance, wavenumber, scale, mean_mu_squared):
    """Factor to multiply log-amplitude (co-)variance with to include log-amplitude saturation.

    ..math:: x = \\frac{1}{1 + r/r_s}
    """
    sat_distance = saturation_distance(mean_mu_squared, wavenumber, scale)
    factor = ( 1.0 / (1.0 + distance/sat_distance) )
    return factor


#def impulse_response_fluctuations(covariance, ntaps, window=None):
    #"""Impulse response describing fluctuations.

    #:param covariance: Covariance vector.
    #:param fs: Sample frequency
    #:param window: Window to apply to impulse response. If passed `None`, no window is applied.
    #:returns: Impulse response of fluctuations filter.

    #"""
    #if window is not None:
        #nsamples = covariance.shape[-1]
        #covariance = covariance * window(nsamples)[...,:] # Not inplace!
    ## The covariance is a symmetric, real function.
    #autospectrum = np.abs(np.fft.rfft(covariance, n=ntaps))#/df**2.0 # Autospectrum
    ##autospectrum[..., 0] = 0.0 # Remove DC component from spectrum.

    ## The autospectrum is real-valued. Taking the square root given an amplitude spectrum.
    ## Because we have a symmetric spectrum, taking the inverse DFT results in an even, real-valued
    ## impulse response. Furthermore, because we have zero phase the impulse response is even as well.
    #ir = np.fft.ifftshift(np.fft.irfft(np.sqrt(autospectrum), n=ntaps).real)

    #return ir


# Uses DCT
def impulse_response_fluctuations(covariance, ntaps, window=None):
    """Impulse response describing fluctuations.

    :param covariance: Covariance vector.
    :param fs: Sample frequency
    :param window: Window to apply to impulse response. If passed `None`, no window is applied.
    :returns: Impulse response of fluctuations filter.

    """
    if window is not None:
        nsamples = covariance.shape[-1]
        covariance = covariance * window(nsamples)[...,:] # Not inplace!
    # The covariance is a symmetric, real function.
    autospectrum = np.abs(dct(covariance, type=1))#/df**2.0 # Autospectrum
    #autospectrum[..., 0] = 0.0 # Remove DC component from spectrum.

    # The autospectrum is real-valued. Taking the square root given an amplitude spectrum.
    # Because we have a symmetric spectrum, taking the inverse DFT results in an even, real-valued
    # impulse response. Furthermore, because we have zero phase the impulse response is even as well.
    ir = np.fft.ifftshift(np.fft.irfft(np.sqrt(autospectrum), n=ntaps).real)

    return ir

def tau(ntaps, fs):
    """Time lag :math:`\\tau` for autocorrelation :math:`B(\\tau)`.

    :param ntaps: Amount of taps.
    :param fs: Sample frequency.

    """
    return np.fft.rfftfreq(ntaps, fs/ntaps)


def correlation_spherical_wave(spatial_separation, correlation_length):
    """Correlation of spherical waves.

    :param spatial_separation: Spatial separation :math:`\\rho`.
    :param correlation_length: Correlation length :math:`\\L_0`.
    :returns: Correlation

    .. math:: \\frac{\\sqrt{\\pi}}{2} \\frac{\\mathrm{erf}{(\\rho/L_0)}}{(\\rho/L_0)}

    .. note:: Instead of spatial separation and correlation length, you can also use time lag and correlation time.

    """
    x = np.atleast_1d(spatial_separation/correlation_length)
    cor = np.sqrt(np.pi) / 2.0 * erf(x)/x
    cor[x==0.0] = 1.0
    return cor


def _correlation_spherical_wave(x):
    cor = np.sqrt(np.pi) / 2.0 * erf(x)/x
    cor[x==0.0] = 1.0
    return cor


def _fluctuations_with_variance(variance_func, fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation):
    """Add correct variance to fluctuations.

    :returns: Fluctuations with correct variance.
    """
    variance = variance_func(distance, wavenumber, correlation_length, mean_mu_squared, include_saturation)
    return fluctuations * np.sqrt(variance)


#def logamp_fluctuations(variance_func, fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation=True):
    #"""Determine log-amplitude fluctuations.

    #:param fluctuations: Fluctuations with variance of one.
    #:param frequency: Frequency to compute fluctuations for.
    #:returns: Log-amplitude fluctuations with correct variance.
    #"""
    #variance_logamp = variance_gaussian(distance, wavenumber, correlation_length, mean_mu_squared, include_saturation=include_saturation)
    #logamp = fluctuations * np.sqrt(variance_logamp)
    #return logamp


#def phase_fluctuations(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared):
    #"""Determine phase fluctuations for given frequency.

    #:param fluctuations: Fluctuations with variance of one.
    #:param frequency: Frequency to compute fluctuations for.
    #:returns: Phase fluctuations with correct variance.
    #"""
    #variance_phase  = variance_gaussian(distance, wavenumber, correlation_length,
                                        #mean_mu_squared, include_saturation=False)
    #phase  = fluctuations * np.sqrt(variance_phase)
    #return phase


def amplitude_fluctuations(logamp):
    """Return amplitude fluctuations given log-amplitude fluctuations.

    :param logamp: Log-amplitude fluctuations.
    :returns: Amplitude fluctuations.

    .. math:: A = \\exp{\\chi}

    """
    return np.exp(logamp)


def delay_fluctuations(phase, fs, frequency=1.0):
    """Return propagation delay fluctuations given phase fluctuations for the specified frequency.

    :param phase: Phase fluctuations for given `frequency`.
    :param frequency: Frequency.
    :returns: Propagation delay fluctuations.

    .. math:: \\mathrm{d}t = -S / (2 \\pi f)

    .. note:: Note the minus sign! This is according to the definition.

    """
    omega = (2.0*np.pi*frequency)
    return phase/omega * (-1)
    #return (phase/omega) / fs * (-1) # We explicitly do a multiplication because Stream does not yet support unary ops


def complex_fluctuations(logamp, phase):
    """Complex fluctuations.

    :param logamp: Log-amplitude fluctuations :math:`\\chi`.
    :param phase: Phase fluctuations :math:`S`.
    :returns: Complex fluctuations :math:`\\Psi`.

    .. math:: \\Psi = e^{\\chi} e^{j\\S}
    """
    return amplitude_fluctuations(logamp) * np.exp(1j*phase)


def transverse_coherence_expected(variance, correlation):
    """Transverse coherence of a spherical waves and Gaussian fluctuations.

    See Daigle, equation 11.
    """
    return np.exp(-2.0*variance * (1.0 - correlation))



def transverse_coherence_expected_large_spatial_separation(variance):
    """Transverse coherence of a spherical waves and Gaussian fluctuations in case the spatial separation is much larger than the correlation length.

    See Daigle, equation 12.
    """
    return np.exp(-2.0 * variance)


def transverse_coherence(logamp_structure, phase_structure):
    """Transverse coherence as function of structure functions.

    See Daigle, equation 6.
    """
    return np.exp(-0.5 * (logamp_structure+phase_structure))


def longitudinal_coherence(logamp_variance, phase_variance):
    """longitudinal coherence.

    See Daigle, equation 13.
    """
    return np.exp(-logamp_variance - phase_variance)


def transverse_speed(velocity_source, orientation):
    """Transverse speed computed from source velocity and source-receiver orientation.

    :param velocity_source: Source velocity.
    :param orientation: Unit vector pointing from source to receiver

    Each row is a sample and each column a spatial dimension.
    The transverse speed is the cross product of the velocity and orientation.

    .. note:: It does not matter whether source-receiver or receiver-source unit vector is given.

    """
    return np.linalg.norm(np.cross(orientation, velocity_source))



"""
Stream
======

Generate a stream of scintillations. Functions in this module are supposed to work with :mod:`streaming.stream`.

Varying correlation time is taken into account by resampling the generated fluctuations.

"""
import streaming
import itertools
from scintillations.common import * # amplitude_fluctuations, delay_fluctuations, impulse_response_fluctuations, variance_gaussian, correlation_spherical_wave
from scintillations.common import _fluctuations_with_variance


def variance_gaussian(distance, wavenumber, scale, mean_mu_squared, include_saturation=False):
    """Variance of Gaussian fluctuations.

    :param distance: Distance.
    :param wavenumber: Wavenumber.
    :param scale: Correlation length
    :param mean_mu_squared: Mean mu squared.

    :param include_saturation: Whether to include log-amplitude saturation. In this case the variance is multiplied with the saturation factor, :func:`saturation_factor`.
    :returns: Variance

    .. math:: \\langle \\chi^2 \\rangle = \\langle S^2 \\rangle = \\frac{\\sqrt{\\pi}}{2} \\langle \\mu^2 \\rangle k^2 r L

    """
    variance = np.sqrt(np.pi)/2.0 * mean_mu_squared.copy() * wavenumber.copy()**2.0 * distance.copy() * scale.copy()
    if include_saturation:
        variance *= saturation_factor(distance, wavenumber, scale, mean_mu_squared)
    return variance


def fluctuations_with_variance(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation=False):
    """Generate fluctuations with correct variance.
    """
    return _fluctuations_with_variance(variance_gaussian, fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, include_saturation)


#from common import *

#def _generate_sequence(ntaps, D, D0, state):

    ## D = v / L * fs

    #x = np.arange(ntaps) #* D0
    #correlation = _correlation_spherical_wave(x)

    ##correlation = correlation_spherical_wave(tau(ntaps, fs_base), fs_base)
    #ir = impulse_response_fluctuations(correlation, ntaps)

    #noise = streaming.signal.noise(state=state)
    ## Fluctuations without taking into account the actual correlation length, speed or sample frequency
    ## Ideal block size? I don't know. Depends on the above parameters!
    #fluctuations = streaming.signal.convolve_overlap_save(noise, streaming.signal.constant(ir), nhop=8192, ntaps=ntaps)

    ## Sample times for the initial sample frequency
    #times_compressed =  streaming.signal.times(D0)

    ## Sample times at output sample frequency, taking into account the correlation time
    #times = streaming.Stream(itertools.chain([0.0], streaming.signal.cumsum(D)))

    #fluctuations = streaming.signal.interpolate(times_compressed, fluctuations, times)
    #return fluctuations


#def generate_fluctuations_resample_fluctuations(ntaps, fs_desired, correlation_time, state, f0):
    #D = 1./(correlation_time * fs_desired)
    #D0 = f0 / f_desired

    #return _generate_sequence(ntaps, D, D0, state)




def generate_fluctuations_resample_fluctuations(ntaps, fs_desired, correlation_time, state, fs_base=50., window=None):
    """Generate fluctuations.

    :param fs_desired: Sample frequency of the input and the output.
    :param fs: Sample frequency at which to compute the initial fluctuations.

    :returns: Fluctuations.

    .. note:: This function applies change of relative speed by resampling the fluctuations at a different pace.

    """
    correlation = correlation_spherical_wave(tau(ntaps, fs_base), fs_base)
    ir = impulse_response_fluctuations(correlation, ntaps)

    # We now generate fluctuations without taking into account the actual correlation length, speed or sample frequency.
    # Ideal block size? I don't know. Depends on the above parameters!
    nhop = ntaps # Arbitrary
    noise = streaming.signal.noise(nblock=nhop, state=state)
    fluctuations = streaming.signal.convolve_overlap_add(noise, streaming.signal.constant(ir), nhop=nhop, ntaps=ntaps)

    # Sample times for the initial sample frequency
    times_compressed = streaming.signal.times(1./fs_base)

    # Sample times at output sample frequency, taking into account the correlation time
    times = streaming.Stream(itertools.chain([0.0], streaming.signal.cumsum(1./correlation_time / fs_desired)))

    # Resample fluctuations to take into account the actual parameters
    fluctuations = streaming.signal.interpolate(times_compressed, fluctuations, times)

    return fluctuations


def generate_fluctuations_resample_filter(fs_desired, fs, ntaps, speed, correlation_length, state, window=None):
    """Generate fluctuations.

    .. note:: This function applies change of relative speed by updating the impulse response of the filter by resampling the initial frequency response.
    """
    return NotImplemented


def generate_fluctuations_update_filter(ntaps, fs, correlation_time, state, window=None):
    """Generate fluctuations.

    .. note:: This function applies change of relative speed by updating the impulse response of the filter by recomputing.
    """
    nblock = ntaps
    _tau = tau(ntaps, fs)
    correlation = correlation_time.map(lambda x: correlation_spherical_wave(_tau, x))

    ir = correlation.map(lambda c: impulse_response_fluctuations(c, ntaps, window=window))
    #ir = streaming.Stream(impulse_response_fluctuations(c, ntaps, window=window) for c in correlation)

    # Noise generator. Samples from the standard normal distribution.
    state = state if state is not None else np.random.RandomState()
    noise = streaming.signal.noise(nblock=None, state=state).blocks(nblock=ntaps, noverlap=ntaps-1)
    #noise = streaming.BlockStream( (state.randn(nblock) for i in itertools.count()), nblock)
    #fluctuations = streaming.Stream(fftconvolve(block, i, mode='valid')[0] for block, i in zip(noise, ir))
    fluctuations = streaming.signal.convolve_overlap_save(noise, ir, nblock, ntaps).samples()
    return fluctuations


def generate_fluctuations(fs_desired, fs, ntaps, speed, correlation_length, state, window=None, method="resample_fluctuations"):
    """Generate fluctuations.
    """
    if method == 'resample_fluctuations':
        return generate_fluctuations_resample_fluctuations(fs_desired, fs, ntaps, speed, correlation_length, state, window=None)
    elif method == 'update_filter':
        return generate_fluctuations_update_filter(fs_desired, fs, ntaps, speed, correlation_length, state, window=None)
    elif method == 'resample_filter':
        return generate_fluctuations_resample_filter(fs_desired, fs, ntaps, speed, correlation_length, state, window=None)



#def fluctuations_naf(nblock, fs, nbins, ntaps, speed, correlation_length,
                     #distance, soundspeed, mean_mu_squared, window, seed, include_saturation):
    #"""Generate gains and delays for amplitude and phase modulations.

    #:param nblock: Blocksize the NAF uses.
    #:param fs: Sample frequency the NAF uses.
    #:param nbins: Amount of frequency bins for the amplitude modulations' spectra.
    #:param ntaps: Amount of points to sample the correlation function.

    #.. note:: Single values per block.
    #"""
    ## Blocks per second in the NAF
    #fs_desired = fs / nblock
    ## Samples per second at which to compute fluctuations
    #fs_low = 1000.

    ## Warn when sample frequency is too low.

    ##------Fluctuations with variance 1---------
    ## Ideally we use the function that works at lower sample frequency. However, then it gets difficult because the
    ## amount of taps is going to be higher than the blocksize (1).
    ##fluctuations = _generate_gaussian_fluctuations(nblock, ntaps, fs, correlation_time, window=window, state=state)
    ## Instead we now consider the upsampled version, partition in blocks with the blocksize the NAF uses,
    ###fluctuations = turbulence_fluctuations(ntaps, fs, fs_low, ntaps, speed,
    ###                                       correlation_length, window=window, state=np.random.RandomState(seed))
    ## and then just pick the last item of the fluctuations.
    ###fluctuations = streaming.Stream(fluctuations.blocks(nblock).map(lambda x: x[-1]))
    #state = np.random.RandomState(seed)
    #fluctuations = generate_fluctuations_resampling(fs_desired, fs_low, ntaps, speed, correlation_length, state, window=window)

    #logamp, phase = fluctuations.tee()

    ##------Amplitude fluctuations--------------
    #frequencies = np.fft.rfftfreq(nbins, 1./fs)
    #wavenumbers = 2.*np.pi*frequencies / soundspeed
    #logamp_func = lambda d : variance_gaussian(d, wavenumbers, correlation_length, mean_mu_squared, include_saturation=include_saturation)
    ## Variances as function of time
    ##variance_logamp = distance.map(logamp_func)
    #variance_logamp = logamp_func(distance)
    ##print(variance_logamp.peek())
    ##print(logamp.peek())

    ## Spectral gains as function of time
    #spectrum = np.exp( logamp * variance_logamp.sqrt() )

    ##------Phase or delay fluctuations---------
    #omega = 2.0*np.pi*1.0
    #wavenumber = omega / soundspeed
    #phase_func = lambda d : variance_gaussian(d, wavenumber, correlation_length, mean_mu_squared)
    #variance_phase = distance.copy().map(phase_func)
    #phase = phase * variance_phase.sqrt()
    #delay = phase / omega

    #return frequencies, spectrum, delay


def _stream_or_constant(x):
    if isinstance(x, streaming.abstractstream.AbstractStream):
        return x.samples()
    else:
        return streaming.signal.constant(x)


def generate_fluctuations_spectra_and_delay(fs, ntaps, correlation_length, speed, frequency, soundspeed,
                                            distance, mean_mu_squared, fmin, include_saturation=True, state=None, window=None):
    """Generate fluctuations.

    :param fs: Sample frequency.
    :param ntaps: Taps of impulse response.
    :param correlation_length: Correlation length.

    :returns: Frequencies, magnitude spectra and delays.
    """
    correlation_length = _stream_or_constant(correlation_length)
    speed = _stream_or_constant(speed)
    soundspeed = _stream_or_constant(soundspeed)
    distance = _stream_or_constant(distance)
    mean_mu_squared = _stream_or_constant(mean_mu_squared)

    correlation_time = correlation_length.copy() / speed.copy()

    fluctuations = generate_fluctuations_resample_fluctuations(ntaps, fs, correlation_time, state, window=window, fs_base=fmin).samples()

    wavenumber_logamp = 2.*np.pi*frequency/soundspeed.copy()
    wavenumber_phase = 2.*np.pi*1.0/soundspeed.copy()
    logamp = fluctuations_with_variance(fluctuations.copy(), wavenumber_logamp, distance.copy(), correlation_length.copy(), mean_mu_squared.copy(), include_saturation)
    phase  = fluctuations_with_variance(fluctuations.copy(), wavenumber_phase , distance.copy(), correlation_length.copy(), mean_mu_squared.copy(), False)

    spectra = amplitude_fluctuations(logamp)
    delay = delay_fluctuations(phase, fs, 1.0)

    del fluctuations, distance, correlation_length, mean_mu_squared, speed, soundspeed

    return spectra, delay


def generate_fluctuations_complex_spectrum(fluctuations):
    """Generate fluctuations.

    :returns: Frequencies, complex spectra
    """
    correlation_length = _stream_or_constant(correlation_length)
    speed = _stream_or_constant(speed)
    soundspeed = _stream_or_constant(soundspeed)
    distance = _stream_or_constant(distance)
    mean_mu_squared = _stream_or_constant(mean_mu_squared)

    correlation_time = correlation_length.copy() / speed

    fluctuations = generate_fluctuations_resample_fluctuations(ntaps, fs, correlation_time, state, window=window, fs_base=fmin).samples()

    wavenumber = 2.*np.pi*frequency/soundspeed
    logamp = fluctuations_with_variance(fluctuations.copy(), wavenumber.copy(), distance.copy(), correlation_length.copy(), mean_mu_squared.copy(), include_saturation)
    phase =  fluctuations_with_variance(fluctuations, wavenumber, distance, correlation_length, mean_mu_squared, False)

    complex_spectra = complex_fluctuations(logamp, phase)
    return complex_spectra


def modulate_with_spectra_and_delay(signal, fs, nhop, spectra, delay, ntaps=None):
    """Modulate signal with amplitude and phase fluctuations.

    :param signal: Carrier signal.
    :param fs: Sample frequency of signal.
    :param nhop: Compute impulse response to describe fluctuations every `nhop` samples.
    :param spectra: Spectrum for each block. Single side magnitude only.
    :param delay: Delay for each block.
    :param ntaps: Amount of filter taps to use for IR of `spectra`.
    :returns: Modulated signal.
    :rtype: :class:`streaming.stream.Stream()`

    .. note:: The spectra describe amplitude fluctuations and the delay is the phase or propagation delay fluctuations.
    """
    #signal = signal.blocks(nhop)

    # Amplitude modulations
    if spectra is not None:
        if ntaps is None:
            raise ValueError("ntaps should be specified when including amplitude modulations.")
        # Calculate an IR for each amplitude modulations spectra
        # We apply linear-phase
        compute_ir = lambda x: np.fft.ifftshift(np.fft.irfft(x, n=ntaps))
        ir = spectra.map(compute_ir)
        #ir = streaming.Stream((np.fft.ifftshift(np.fft.irfft(2.*x, n=ntaps)) for x in spectra))

        # Apply amplitude modulations
        signal = streaming.signal.convolve_overlap_save(signal, ir, nhop, ntaps)
        # Linear-phase filter. Correct for group delay.
        signal = signal.samples().drop(int(ntaps//2))

    # Time-delay modulations
    if delay is not None:
        # Interpolate time-delay modulations
        times = streaming.signal.times(1./fs)
        fb = fs / nhop
        times_low = streaming.signal.times(1./fb)
        delay = streaming.signal.interpolate(times_low, delay, times.copy())
        # Apply time-delay modulations
        signal = streaming.signal.vdl(signal, times, delay)

    return signal


def modulate_with_complex_spectra(signal, fs, nhop, complex_spectra, ntaps):
    """Modulate signal with amplitude and phase fluctuations.

    .. note:: The complex spectra includes amplitude and phase fluctuations.

    """
    ir = spectrum.map(lambda x: np.fft.ifftshift(np.fft.irfft(x, n=ntaps)))
    # Convolve signal with complex spectra.
    # Because we add a linear-phase (ifftshift) we correct for that by dropping samples.
    signal = streaming.signal.convolve_overlap_save(signal, ir, nhop, ntaps).samples().drop(ntaps//2)
    return signal


#def apply_modulations(signal, fs, nblock, spectrum, delay, ntaps=None):
    #"""Modulate `signal` with amplitude modulations given by `spectrum` and time-delay modulations given by `delay`.

    #"""
    #return modulate_with_complex_spectra(signal, fs, nblock, spectrum, delay, ntaps=ntaps)


def modulate(signal, fs, nhop, correlation_length, speed, distance, soundspeed, mean_mu_squared, fmin, ntaps_corr=8192,
             ntaps_spectra=128, window=None, include_saturation=False, state=None,
             include_amplitude=True, include_phase=True, method="spectra_and_delay"):
    """Compute and apply scintillations to `signal`.

    :param signal: Signal to modulate.
    :type signal: :class:`Stream`
    :param fs: Sample frequency
    :param nhop: Compute new impulse response and delay every `nhop` samples.
    :param correlation: Correlation length.
    :type correlation: Either a single value, or a :class:`Stream` sampled at `fs/nhop`.
    :param speed: Transverse speed.
    :type correlation: Either a single value, or a :class:`Stream` sampled at `fs/nhop`.
    :param distance: Source-receiver distance.
    :type distance: Either a single value, or a :class:`Stream` sampled at `fs/nhop`.
    :param soundspeed: Speed of sound.
    :type soundspeed: Either a single value, or a :class:`Stream` sampled at `fs/nhop`.
    :param mean_mu_squared: Variance of the refractive-index field.
    :type mean_mu_squared: Either a single value, or a :class:`Stream` sampled at `fs/nhop`.
    :param ntaps_corr: Length of the impulse responses for the correlation filter.
    :param ntaps_spectra: Length of the impulse responses for the eventual magnitude-filter.
    :param window: Window
    :param include_saturation: Whether to include log-amplitude saturation.
    :param state: State of the PRNG.
    :param include_amplitude: Whether to include log-amplitude fluctuations.
    :param include_phase: Whether to include phase fluctuations.
    :param method: Method to use.


    """
    # Generate complex spectra or magnitude spectra and delays
    # Block rate
    fb = fs / nhop
    frequency = np.fft.rfftfreq(ntaps_spectra, 1./ fs)

    ## Downsample speed, correlation_length, distance, soundspeed, mean_mu_squared
    ## TODO: lowpass filter?
    #if isinstance(speed, streaming.abstractstream.AbstractStream):               speed = speed.samples().take_nth(nhop)
    #if isinstance(correlation_length, streaming.abstractstream.AbstractStream):  correlation_length = correlation_length.samples().take_nth(nhop)
    #if isinstance(distance, streaming.abstractstream.AbstractStream):            distance = distance.samples().take_nth(nhop)
    #if isinstance(soundspeed, streaming.abstractstream.AbstractStream):          soundspeed = soundspeed.samples().take_nth(nhop)
    #if isinstance(mean_mu_squared, streaming.abstractstream.AbstractStream):     mean_mu_squared = mean_mu_squared.samples().take_nth(nhop)

    if method=="complex_spectra":
        complex_spectra = generate_fluctuations_complex_spectra(fb, ntaps_corr, correlation_length, speed,
                                                                frequency, soundspeed, distance, mean_mu_squared,
                                                                fmin, include_saturation, state, window)

        modulated = modulate_with_complex_spectra(signal, fs, nhop, complex_spectra, ntaps=ntaps_spectra)

    elif method=="spectra_and_delay":

        spectra, delay = generate_fluctuations_spectra_and_delay(fb, ntaps_corr, correlation_length, speed,
                                                                 frequency, soundspeed, distance, mean_mu_squared,
                                                                 fmin, include_saturation, state, window)
        # Apply modulations
        if not include_amplitude:
            spectra = None
        if not include_phase:
            delay = None
        modulated = modulate_with_spectra_and_delay(signal, fs, nhop, spectra, delay, ntaps=ntaps_spectra)
    else:
        raise ValueError("Incorrect method.")

    return modulated







"""
==============
Scintillations
==============

Atmospheric turbulence causes fluctuations in the sound speed which in effects causes fluctuations in the amplitude and phase of the sound pressure.
This module contains functions for generating and analyzing sequences of scintillations that follow a Gaussian spectrum,
"""

import scintillations.common
import scintillations.sequence
import scintillations.stream



import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import numpy as np


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

CLASSIFIERS = [
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.5',
    'Topic :: Scientific/Engineering',
    ]

setup(
    name='scintillations',
    version='0.1.0',
    description="Generate sequences of scintillations",
    author='Frederik Rietdijk',
    author_email='freddyrietdijk@fridh.nl',
    license='LICENSE',
    packages=find_packages(exclude=["tests"]),
    zip_safe=False,
    install_requires=[
        'numpy',
        'scipy',
        'streaming',
        ],
    classifiers=CLASSIFIERS,
    tests_require = [ 'pytest', 'acoustics' ],
    cmdclass = {'test': PyTest},
    )
